@@ -29,6 +29,7 @@ enum
 {
   WINDOW_OPENED,
   WINDOW_CLOSED,
+  STACKING_CHANGED,
   ACTIVE_WINDOW_CHANGED,
 
   LAST_SIGNAL,
@@ -62,6 +63,7 @@ on_state_file_load_timeout (BamfLegacyScreen *self)
   gchar *line, *name, *class, *exec;
   GList *l;
   gchar **parts;
+  gsize parts_size;
   guint32 xid;
 
   g_return_val_if_fail (BAMF_IS_LEGACY_SCREEN (self), FALSE);
@@ -78,26 +80,36 @@ on_state_file_load_timeout (BamfLegacyScreen *self)
   // close	<xid>
   // attention	<xid>	<true/false>
   // skip	<xid>	<true/false>
+  // geometry <xid> <x> <y> <width> <height>
+  // maximized <xid> <maximized/vmaximized/hmaximized/floating>
 
   parts = g_strsplit (line, "\t", 0);
   g_free (line);
 
+  parts_size = 0;
+  while (parts[parts_size] != NULL)
+    parts_size++;
+
+  if (parts_size < 2)
+    return FALSE;
+
   xid = (guint32) atol (parts[1]);
-  if (g_strcmp0 (parts[0], "open") == 0)
+  if (g_strcmp0 (parts[0], "open") == 0 && parts_size == 5)
     {
       name  = parts[2];
       class = parts[3];
       exec  = parts[4];
 
       window = BAMF_LEGACY_WINDOW (bamf_legacy_window_test_new (xid, name, class, exec));
-      self->priv->windows = g_list_prepend (self->priv->windows, window);
+      self->priv->windows = g_list_append (self->priv->windows, window);
+      g_signal_emit (window, legacy_screen_signals[STACKING_CHANGED], 0);
 
       g_signal_connect (G_OBJECT (window), "closed",
                         (GCallback) handle_window_closed, self);
 
       g_signal_emit (self, legacy_screen_signals[WINDOW_OPENED], 0, window);
     }
-  else if (g_strcmp0 (parts[0], "close") == 0)
+  else if (g_strcmp0 (parts[0], "close") == 0 && parts_size == 2)
     {
       for (l = self->priv->windows; l; l = l->next)
         {
@@ -109,7 +121,7 @@ on_state_file_load_timeout (BamfLegacyScreen *self)
             }
         }
     }
-  else if (g_strcmp0 (parts[0], "attention") == 0)
+  else if (g_strcmp0 (parts[0], "attention") == 0 && parts_size == 3)
     {
       gboolean attention = FALSE;
       if (g_strcmp0 (parts[2], "true") == 0)
@@ -128,7 +140,7 @@ on_state_file_load_timeout (BamfLegacyScreen *self)
             }
         }
     }
-  else if (g_strcmp0 (parts[0], "skip") == 0)
+  else if (g_strcmp0 (parts[0], "skip") == 0 && parts_size ==  3)
     {
       gboolean skip = FALSE;
       if (g_strcmp0 (parts[2], "true") == 0)
@@ -147,6 +159,46 @@ on_state_file_load_timeout (BamfLegacyScreen *self)
             }
         }
     }
+  else if (g_strcmp0 (parts[0], "geometry") == 0 && parts_size == 6)
+    {
+      int x = atoi (parts[2]);
+      int y = atoi (parts[3]);
+      int width = atoi (parts[4]);
+      int height = atoi (parts[5]);
+
+      for (l = self->priv->windows; l; l = l->next)
+        {
+          if (bamf_legacy_window_get_xid (l->data) == xid)
+            {
+              bamf_legacy_window_test_set_geometry (l->data, x, y, width, height);
+              break;
+            }
+        }
+    }
+  else if (g_strcmp0 (parts[0], "maximized") == 0 && parts_size == 3)
+    {
+      BamfWindowMaximizationType maximized;
+
+      if (g_strcmp0 (parts[2], "maximized") == 0)
+        maximized = BAMF_WINDOW_MAXIMIZED;
+      else if (g_strcmp0 (parts[2], "vmaximized") == 0)
+        maximized = BAMF_WINDOW_VERTICAL_MAXIMIZED;
+      else if (g_strcmp0 (parts[2], "hmaximized") == 0)
+        maximized = BAMF_WINDOW_HORIZONTAL_MAXIMIZED;
+      else if (g_strcmp0 (parts[2], "floating") == 0)
+        maximized = BAMF_WINDOW_FLOATING;
+      else
+        return TRUE;
+
+      for (l = self->priv->windows; l; l = l->next)
+        {
+          if (bamf_legacy_window_get_xid (l->data) == xid)
+            {
+              bamf_legacy_window_test_set_maximized (l->data, maximized);
+              break;
+            }
+        }
+    }
   else
     {
       g_warning ("Could not parse line\n");
@@ -156,21 +208,110 @@ on_state_file_load_timeout (BamfLegacyScreen *self)
   return TRUE;
 }
 
+static gint
+compare_windows_by_stack_order (gconstpointer a, gconstpointer b, gpointer data)
+{
+  BamfLegacyScreen *self;
+  GList *l;
+  guint xid_a, xid_b;
+  guint idx_a, idx_b;
+
+  g_return_val_if_fail (BAMF_IS_LEGACY_SCREEN (data), 1);
+  self = BAMF_LEGACY_SCREEN (data);
+
+  xid_a = bamf_legacy_window_get_xid (BAMF_LEGACY_WINDOW (a));
+  xid_b = bamf_legacy_window_get_xid (BAMF_LEGACY_WINDOW (b));
+
+  gboolean idx_a_found = FALSE;
+  gboolean idx_b_found = FALSE;
+  idx_a = 0;
+  idx_b = 0;
+
+  for (l = wnck_screen_get_windows_stacked (self->priv->legacy_screen); l; l = l->next)
+  {
+    gulong legacy_xid = wnck_window_get_xid (WNCK_WINDOW (l->data));
+
+    if (!idx_a_found)
+      {
+        if (xid_a != legacy_xid)
+          idx_a++;
+        else
+          idx_a_found = TRUE;
+      }
+
+    if (!idx_b_found)
+      {
+        if (xid_b != legacy_xid)
+          idx_b++;
+        else
+          idx_b_found = TRUE;
+      }
+
+    if (idx_a_found && idx_b_found)
+      break;
+  }
+
+  return (idx_a < idx_b) ? -1 : 1;
+}
+
 static void
 handle_window_opened (WnckScreen *screen, WnckWindow *window, BamfLegacyScreen *legacy)
 {
   BamfLegacyWindow *legacy_window;
+  g_return_if_fail (WNCK_IS_WINDOW (window));
 
   legacy_window = bamf_legacy_window_new (window);
 
   g_signal_connect (G_OBJECT (legacy_window), "closed",
                     (GCallback) handle_window_closed, legacy);
 
-  legacy->priv->windows = g_list_prepend (legacy->priv->windows, legacy_window);
+  legacy->priv->windows = g_list_insert_sorted_with_data (legacy->priv->windows, legacy_window, 
+                                                          compare_windows_by_stack_order,
+                                                          legacy);
 
   g_signal_emit (legacy, legacy_screen_signals[WINDOW_OPENED], 0, legacy_window);
 }
 
+static void
+handle_stacking_changed (WnckScreen *screen, BamfLegacyScreen *legacy)
+{
+  legacy->priv->windows = g_list_sort_with_data (legacy->priv->windows,
+                                                 compare_windows_by_stack_order,
+                                                 legacy);
+
+  g_signal_emit (legacy, legacy_screen_signals[STACKING_CHANGED], 0);
+}
+
+/* This function allows to push into the screen a window by its xid.
+ * If the window is already known, it's just ignored, otherwise it gets added
+ * to the windows list. The BamfLegacyScreen should automatically update its
+ * windows list when they are added/removed from the screen, but if a child
+ * BamfLegacyWindow is closed, then it could be possible to re-add it.        */
+void
+bamf_legacy_screen_inject_window (BamfLegacyScreen *self, guint xid)
+{
+  g_return_if_fail (BAMF_IS_LEGACY_SCREEN (self));
+  BamfLegacyWindow *window;
+  GList *l;
+
+  for (l = self->priv->windows; l; l = l->next)
+    {
+      window = l->data;
+
+      if (bamf_legacy_window_get_xid (window) == xid)
+        {
+          return;
+        }
+    }
+
+  WnckWindow *legacy_window = wnck_window_get(xid);
+
+  if (WNCK_IS_WINDOW (legacy_window))
+    {
+      handle_window_opened(NULL, legacy_window, self);
+    }
+}
+
 void
 bamf_legacy_screen_set_state_file (BamfLegacyScreen *self,
                                    const char *file)
@@ -183,6 +324,7 @@ bamf_legacy_screen_set_state_file (BamfLegacyScreen *self,
   // Disconnect our handlers so we can work purely on the file
   g_signal_handlers_disconnect_by_func (self->priv->legacy_screen, handle_window_opened, self);
   g_signal_handlers_disconnect_by_func (self->priv->legacy_screen, handle_window_closed, self);
+  g_signal_handlers_disconnect_by_func (self->priv->legacy_screen, handle_stacking_changed, self);
 
   gfile = g_file_new_for_path (file);
 
@@ -261,33 +403,42 @@ bamf_legacy_screen_class_init (BamfLegacyScreenClass * klass)
   g_type_class_add_private (klass, sizeof (BamfLegacyScreenPrivate));
 
   legacy_screen_signals [WINDOW_OPENED] =
-  	g_signal_new ("window-opened",
-  	              G_OBJECT_CLASS_TYPE (klass),
-  	              G_SIGNAL_RUN_FIRST,
-  	              G_STRUCT_OFFSET (BamfLegacyScreenClass, window_opened),
-  	              NULL, NULL,
-  	              g_cclosure_marshal_VOID__OBJECT,
-  	              G_TYPE_NONE, 1,
-  	              BAMF_TYPE_LEGACY_WINDOW);
+    g_signal_new (BAMF_LEGACY_SCREEN_SIGNAL_WINDOW_OPENED,
+                  G_OBJECT_CLASS_TYPE (klass),
+                  G_SIGNAL_RUN_FIRST,
+                  G_STRUCT_OFFSET (BamfLegacyScreenClass, window_opened),
+                  NULL, NULL,
+                  g_cclosure_marshal_VOID__OBJECT,
+                  G_TYPE_NONE, 1,
+                  BAMF_TYPE_LEGACY_WINDOW);
 
   legacy_screen_signals [WINDOW_CLOSED] =
-  	g_signal_new ("window-closed",
-  	              G_OBJECT_CLASS_TYPE (klass),
-  	              G_SIGNAL_RUN_FIRST,
-  	              G_STRUCT_OFFSET (BamfLegacyScreenClass, window_closed),
-  	              NULL, NULL,
-  	              g_cclosure_marshal_VOID__OBJECT,
-  	              G_TYPE_NONE, 1,
-  	              BAMF_TYPE_LEGACY_WINDOW);
+    g_signal_new (BAMF_LEGACY_SCREEN_SIGNAL_WINDOW_CLOSED,
+                  G_OBJECT_CLASS_TYPE (klass),
+                  G_SIGNAL_RUN_FIRST,
+                  G_STRUCT_OFFSET (BamfLegacyScreenClass, window_closed),
+                  NULL, NULL,
+                  g_cclosure_marshal_VOID__OBJECT,
+                  G_TYPE_NONE, 1,
+                  BAMF_TYPE_LEGACY_WINDOW);
+
+  legacy_screen_signals [STACKING_CHANGED] =
+    g_signal_new (BAMF_LEGACY_SCREEN_SIGNAL_STACKING_CHANGED,
+                  G_OBJECT_CLASS_TYPE (klass),
+                  G_SIGNAL_RUN_FIRST,
+                  G_STRUCT_OFFSET (BamfLegacyScreenClass, stacking_changed),
+                  NULL, NULL,
+                  g_cclosure_marshal_VOID__VOID,
+                  G_TYPE_NONE, 0);
 
   legacy_screen_signals [ACTIVE_WINDOW_CHANGED] =
-  	g_signal_new ("active-window-changed",
-  	              G_OBJECT_CLASS_TYPE (klass),
-  	              G_SIGNAL_RUN_FIRST,
-  	              G_STRUCT_OFFSET (BamfLegacyScreenClass, active_window_changed),
-  	              NULL, NULL,
-  	              g_cclosure_marshal_VOID__VOID,
-  	              G_TYPE_NONE, 0);
+    g_signal_new (BAMF_LEGACY_SCREEN_SIGNAL_ACTIVE_WINDOW_CHANGED,
+                  G_OBJECT_CLASS_TYPE (klass),
+                  G_SIGNAL_RUN_FIRST,
+                  G_STRUCT_OFFSET (BamfLegacyScreenClass, active_window_changed),
+                  NULL, NULL,
+                  g_cclosure_marshal_VOID__VOID,
+                  G_TYPE_NONE, 0);
 }
 
 static BamfLegacyScreen *self = NULL;
@@ -305,6 +456,9 @@ bamf_legacy_screen_get_default ()
   g_signal_connect (G_OBJECT (self->priv->legacy_screen), "window-opened",
                     (GCallback) handle_window_opened, self);
 
+  g_signal_connect (G_OBJECT (self->priv->legacy_screen), "window-stacking-changed",
+                    (GCallback) handle_stacking_changed, self);
+
   g_signal_connect (G_OBJECT (self->priv->legacy_screen), "active-window-changed",
                     (GCallback) handle_active_window_changed, self);
 
