@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2010-2011 Canonical Ltd
+ * Copyright (C) 2010-2012 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -28,6 +28,7 @@
 #include "bamf-legacy-screen.h"
 #include "bamf-indicator-source.h"
 #include "bamf-xutils.h"
+#include <strings.h>
 
 G_DEFINE_TYPE (BamfMatcher, bamf_matcher, BAMF_DBUS_TYPE_MATCHER_SKELETON);
 #define BAMF_MATCHER_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE(obj, \
@@ -126,6 +127,37 @@ on_view_active_changed (BamfView *view, gboolean active, BamfMatcher *matcher)
 
 static void bamf_matcher_unregister_view (BamfMatcher *self, BamfView *view);
 
+static BamfApplication *
+bamf_matcher_get_application_by_desktop_file (BamfMatcher *self, const char *desktop_file)
+{
+  GList *l;
+  BamfView *view;
+
+  g_return_val_if_fail (BAMF_IS_MATCHER (self), NULL);
+
+  if (!desktop_file)
+    return NULL;
+
+  for (l = self->priv->views; l; l = l->next)
+    {
+      view = l->data;
+
+      if (!BAMF_IS_APPLICATION (view))
+        continue;
+
+      BamfApplication *app = BAMF_APPLICATION (view);
+      const gchar *app_desktop;
+      app_desktop = bamf_application_get_desktop_file (app);
+
+      if (g_strcmp0 (desktop_file, app_desktop) == 0)
+        {
+          return app;
+        }
+    }
+
+  return NULL;
+}
+
 static gboolean
 emit_paths_changed (gpointer user_data)
 {
@@ -177,11 +209,8 @@ emit_paths_changed (gpointer user_data)
 
 static void bamf_matcher_prepare_path_change (BamfMatcher *self, const gchar *desktop_file, ViewChangeType change_type)
 {
-  GList *l;
-  BamfView *view;
   BamfMatcherPrivate *priv;
-  const char *app_desktop;
-  gboolean found = FALSE;
+  BamfApplication *app;
 
   if (desktop_file == NULL) return;
 
@@ -189,26 +218,14 @@ static void bamf_matcher_prepare_path_change (BamfMatcher *self, const gchar *de
 
   priv = self->priv;
 
-  for (l = priv->views; l; l = l->next)
-    {
-      view = l->data;
-
-      if (!BAMF_IS_APPLICATION (view) || !bamf_view_is_running (view))
-        continue;
-
-      app_desktop = bamf_application_get_desktop_file (BAMF_APPLICATION (view));
-      if (!app_desktop) continue;
-
-      if (g_strcmp0 (desktop_file, app_desktop) == 0)
-        {
-          found = TRUE;
-          break;
-        }
-    }
-
   /* the app was already running (ADDED) / had more instances which are still
    * there (REMOVED) */
-  if (found) return;
+  app = bamf_matcher_get_application_by_desktop_file (self, desktop_file);
+
+  if (BAMF_IS_APPLICATION (app) && bamf_view_is_running (BAMF_VIEW (app)))
+    {
+      return;
+    }
 
   if (!priv->opened_closed_paths_table)
     {
@@ -258,7 +275,7 @@ bamf_matcher_register_view_stealing_ref (BamfMatcher *self, BamfView *view)
   self->priv->views = g_list_prepend (self->priv->views, view);
 
   g_signal_emit_by_name (self, "view-opened", path, type);
-  
+
   // trigger manually since this is already active
   if (bamf_view_is_active (view))
     on_view_active_changed (view, TRUE, self);
@@ -642,9 +659,11 @@ pid_parent_tree (BamfMatcher *self, gint pid)
 }
 
 static gboolean
-exec_string_should_be_processed (BamfMatcher *self,
-                                 char *exec)
+exec_string_should_be_processed (const char *exec)
 {
+  if (!exec)
+    return TRUE;
+
   return !g_str_has_prefix (exec, "ooffice") && !g_str_has_prefix (exec, "libreoffice");
 }
 
@@ -812,7 +831,7 @@ load_desktop_file_to_table (BamfMatcher * self,
       return;
     }
 
-  if (exec_string_should_be_processed (self, exec))
+  if (exec_string_should_be_processed (exec))
     {
       /**
        * Set of nasty hacks which should be removed some day. We wish to keep the full exec
@@ -922,7 +941,7 @@ load_index_file_to_table (BamfMatcher * self,
       gchar **parts = g_strsplit (line, "\t", 3);
       exec = parts[1];
 
-      if (exec_string_should_be_processed (self, exec))
+      if (exec_string_should_be_processed (exec))
         {
           char *tmp = trim_exec_string (self, exec);
           g_free (parts[1]);
@@ -1360,6 +1379,9 @@ is_open_office_window (BamfMatcher * self, BamfLegacyWindow * window)
 {
   const char *class_name = bamf_legacy_window_get_class_name (window);
 
+  if (!class_name)
+    return FALSE;
+
   return (g_str_has_prefix (class_name, "LibreOffice") ||
           g_str_has_prefix (class_name, "libreoffice") ||
           g_str_has_prefix (class_name, "OpenOffice") ||
@@ -1514,59 +1536,158 @@ static gboolean
 is_web_app_window (BamfMatcher *self, BamfLegacyWindow *window)
 {
   const char *window_class = bamf_legacy_window_get_class_name (window);
-#ifdef USE_GTK3
-  const char *instance_name = bamf_legacy_window_get_class_instance_name(window);
-  // chromium uses url wm_class strings to represent its web apps. These apps
-  // will still have the same parent pid and hints as the main chrome window, so we
-  // skip the hint check. We can tell a window is a chrome web app window if its instance
-  // name is google-chrome but its window class is not Google-chrome
-  return !g_strcmp0(instance_name, "google-chrome") && g_strcmp0(window_class, "Google-chrome");
-#else
-  return g_str_has_prefix(window_class, "crx_");
-#endif
+  const char *instance_name = bamf_legacy_window_get_class_instance_name (window);
+
+  // Chrome/Chromium uses url wm_class strings to represent its web apps.
+  // These apps will still have the same parent pid and hints as the main chrome
+  // window, so we skip the hint check.
+  // We can tell a window is a chrome web app window if its instance name is
+  // not google-chrome but its window class is Google-chrome
+  // We can tell a window is chromium web app window if its instance name is
+  // not chromium-browser but its window class is Chromium Browser
+
+  gboolean valid_app = FALSE;
+
+  if (instance_name && window_class)
+    {
+      if (g_strcmp0 (window_class, "Google-chrome") == 0 &&
+          g_strcmp0 (instance_name, "google-chrome") != 0 &&
+          !g_str_has_prefix (instance_name, "Google-chrome"))
+        {
+          valid_app = TRUE;
+        }
+      else if (g_strcmp0 (window_class, "Chromium-browser") == 0 &&
+               g_strcmp0 (instance_name, "chromium-browser") != 0 &&
+               !g_str_has_prefix (instance_name, "Chromium-browser"))
+        {
+          valid_app = TRUE;
+        }
+    }
+
+  return valid_app;
+}
+
+static gboolean
+bamf_matcher_window_skips_hint_set (BamfMatcher *self, BamfLegacyWindow *window)
+{
+  gboolean skip_hint_set = FALSE;
+  g_return_val_if_fail (BAMF_IS_MATCHER (self), TRUE);
+
+  skip_hint_set = is_web_app_window (self, window);
+
+  return skip_hint_set;
+}
+
+static GList *
+bamf_matcher_get_class_matching_desktop_files (BamfMatcher *self, const gchar *class_name)
+{
+  GList* desktop_files = NULL;
+  gpointer key;
+  gpointer value;
+  GHashTableIter iter;
+
+  g_return_val_if_fail (BAMF_IS_MATCHER (self), NULL);
+  g_hash_table_iter_init (&iter, self->priv->desktop_class_table);
+
+  while (g_hash_table_iter_next (&iter, &key, &value))
+    {
+      gchar* desktop_file = g_strdup (key);
+      gchar* desktop_class = value;
+
+      if (g_strcmp0 (desktop_class, class_name) == 0)
+        {
+          desktop_files = g_list_prepend (desktop_files, desktop_file);
+        }
+    }
+
+  return desktop_files;
+}
+
+static gboolean
+bamf_matcher_has_instance_class_desktop_file (BamfMatcher *self, const gchar *class_name)
+{
+  gpointer key;
+  gpointer value;
+  GHashTableIter iter;
+
+  g_return_val_if_fail (BAMF_IS_MATCHER (self), FALSE);
+  g_hash_table_iter_init (&iter, self->priv->desktop_class_table);
+
+  while (g_hash_table_iter_next (&iter, &key, &value))
+    {
+      gchar* desktop_class = value;
+
+      if (g_strcmp0 (desktop_class, class_name) == 0)
+        {
+          return TRUE;
+        }
+    }
+
+  return FALSE;
 }
 
 static GList *
 bamf_matcher_possible_applications_for_window (BamfMatcher *self,
-                                               BamfWindow *bamf_window)
+                                               BamfWindow *bamf_window,
+                                               const char **target_class_out)
 {
-  char *hint = NULL;
   BamfMatcherPrivate *priv;
-  gint pid;
   BamfLegacyWindow *window;
-  GList *desktop_files = NULL, *pid_list = NULL, *l;
+  GList *desktop_files = NULL, *l;
+  char *desktop_file = NULL;
+  const char *desktop_class = NULL;
+  const char *class_name = NULL;
+  const char *instance_name = NULL;
+  const char *target_class = NULL;
+  gboolean filter_by_wmclass = FALSE;
 
   g_return_val_if_fail (BAMF_IS_WINDOW (bamf_window), NULL);
   g_return_val_if_fail (BAMF_IS_MATCHER (self), NULL);
 
   priv = self->priv;
-
   window = bamf_window_get_window (bamf_window);
+  desktop_file = get_window_hint (window, _NET_WM_DESKTOP_FILE);
+  class_name = bamf_legacy_window_get_class_name (window);
+  instance_name = bamf_legacy_window_get_class_instance_name (window);
 
-  hint = get_window_hint (window, _NET_WM_DESKTOP_FILE);
-  const char *window_class = bamf_legacy_window_get_class_name (window);
+  target_class = instance_name;
+  filter_by_wmclass = bamf_matcher_has_instance_class_desktop_file (self, target_class);
 
-  if (hint)
+  if (!filter_by_wmclass)
+  {
+    if (is_web_app_window (self, window))
+      {
+        // This ensures that a new application is created even for unknown webapps
+        filter_by_wmclass = TRUE;
+      }
+    else
+      {
+        target_class = class_name;
+        filter_by_wmclass = bamf_matcher_has_instance_class_desktop_file (self, target_class);
+      }
+  }
+
+  if (desktop_file)
     {
-      /* whew, hard work, didn't even have to make a copy! */
-      if (hint[0] != '\0' && !is_web_app_window(self, window))
+      desktop_class = bamf_matcher_get_desktop_file_class (self, desktop_file);
+
+      if ((!filter_by_wmclass && !desktop_class) || g_strcmp0 (desktop_class, target_class) == 0)
         {
-          desktop_files = g_list_prepend (desktop_files, hint);
+          desktop_files = g_list_prepend (desktop_files, desktop_file);
         }
       else
         {
-          g_free (hint);
+          g_free (desktop_file);
         }
     }
-  else
+
+  desktop_file = NULL;
+
+  if (!desktop_files)
     {
-      
-      char *desktop_file;
-      char *desktop_class;
-      
-      if (window_class)
+      if (class_name)
         {
-          char *window_class_down = g_ascii_strdown (window_class, -1);
+          char *window_class_down = g_ascii_strdown (class_name, -1);
           l = g_hash_table_lookup (priv->desktop_id_table, window_class_down);
           g_free (window_class_down);
 
@@ -1576,12 +1697,15 @@ bamf_matcher_possible_applications_for_window (BamfMatcher *self,
 
               if (desktop_file)
                 {
-                  desktop_class = g_hash_table_lookup (priv->desktop_class_table, desktop_file);
-                  if ((desktop_class == NULL || g_strcmp0 (desktop_class, window_class) == 0) &&
-                      !g_list_find_custom (desktop_files, desktop_file,
-                                           (GCompareFunc) g_strcmp0))
+                  desktop_class = bamf_matcher_get_desktop_file_class (self, desktop_file);
+
+                  if ((!filter_by_wmclass && !desktop_class) || g_strcmp0 (desktop_class, target_class) == 0)
                     {
-                      desktop_files = g_list_prepend (desktop_files, g_strdup (desktop_file));
+                      if (!g_list_find_custom (desktop_files, desktop_file,
+                                               (GCompareFunc) g_strcmp0))
+                        {
+                          desktop_files = g_list_prepend (desktop_files, g_strdup (desktop_file));
+                        }
                     }
                 }
             }
@@ -1589,31 +1713,10 @@ bamf_matcher_possible_applications_for_window (BamfMatcher *self,
           desktop_files = g_list_reverse (desktop_files);
         }
 
-      /* Iterate over the desktop class table, and add matching desktop files */
-      gpointer key;
-      gpointer value;
-      GHashTableIter iter;
-      g_hash_table_iter_init (&iter, priv->desktop_class_table);
-
-      while (g_hash_table_iter_next (&iter, &key, &value))
-        {
-          desktop_file = g_strdup (key);
-          desktop_class = value;
-          if (g_strcmp0 (desktop_class, window_class) == 0 &&
-              !g_list_find_custom (desktop_files, desktop_file, (GCompareFunc) g_strcmp0))
-            {
-              desktop_files = g_list_prepend (desktop_files, desktop_file);
-            }
-          else
-            {
-              g_free (desktop_file);
-            }
-        }
-
-      pid = bamf_legacy_window_get_pid (window);
-      pid_list = bamf_matcher_possible_applications_for_pid (self, pid);
+      gint pid = bamf_legacy_window_get_pid (window);
+      GList *pid_list = bamf_matcher_possible_applications_for_pid (self, pid);
       
-      /* Append these files to the end to give preference to window_class style picking.
+      /* Append these files to the end to give preference to class_name style picking.
          This style of matching is prefered and used by GNOME Shell however does not work
          very well in practice, thus requiring the fallback here */
       for (l = pid_list; l; l = l->next)
@@ -1627,10 +1730,10 @@ bamf_matcher_possible_applications_for_window (BamfMatcher *self,
             {
               gboolean append = FALSE;
 
-              if (window_class)
+              if (target_class)
                 {
-                  desktop_class = g_hash_table_lookup (priv->desktop_class_table, desktop_file);
-                  if (desktop_class == NULL || g_strcmp0 (desktop_class, window_class) == 0)
+                  desktop_class = bamf_matcher_get_desktop_file_class (self, desktop_file);
+                  if ((!filter_by_wmclass && !desktop_class) || g_strcmp0 (desktop_class, target_class) == 0)
                     {
                       append = TRUE;
                     }
@@ -1659,7 +1762,7 @@ bamf_matcher_possible_applications_for_window (BamfMatcher *self,
                             }
                         }
                     }
-                  
+
                   desktop_files = g_list_insert_before (desktop_files, last, desktop_file);
                 }
               else
@@ -1671,85 +1774,169 @@ bamf_matcher_possible_applications_for_window (BamfMatcher *self,
 
       g_list_free (pid_list);
     }
-  
+
+  if (!desktop_files && filter_by_wmclass)
+    {
+      desktop_files = bamf_matcher_get_class_matching_desktop_files (self, target_class);
+    }
+
+  if (target_class_out)
+    {
+      *target_class_out = target_class;
+    }
+
   return desktop_files;
 }
 
-static void
-bamf_matcher_setup_window_state (BamfMatcher *self,
-                                 BamfWindow *bamf_window)
+static BamfApplication *
+bamf_matcher_get_application_for_window (BamfMatcher *self,
+                                         BamfWindow *bamf_window,
+                                         gboolean *new_application)
 {
   GList *possible_apps, *l;
   BamfLegacyWindow *window;
-  GList *views, *a;
-  const char *win_class;
-  const char *desktop_file;
-  const char *app_class;
+  const gchar *win_class_name;
+  const gchar *target_class = NULL;
+  const gchar *app_class = NULL;
+  const gchar *app_desktop = NULL;
   BamfApplication *app = NULL, *best = NULL;
-  BamfView *view;
 
-  g_return_if_fail (BAMF_IS_MATCHER (self));
-  g_return_if_fail (BAMF_IS_WINDOW (bamf_window));
+  g_return_val_if_fail (BAMF_IS_MATCHER (self), NULL);
+  g_return_val_if_fail (BAMF_IS_WINDOW (bamf_window), NULL);
 
   window = bamf_window_get_window (bamf_window);
-  views = self->priv->views;
+  win_class_name = bamf_legacy_window_get_class_name (window);
 
-  possible_apps = bamf_matcher_possible_applications_for_window (self, bamf_window);
-  win_class = bamf_legacy_window_get_class_name (window);
+  possible_apps = bamf_matcher_possible_applications_for_window (self, bamf_window, &target_class);
+  app_class = target_class;
 
-  /* Loop over every application, inside that application see if its .desktop file
-   * matches with any of our possible hits. If so we match it. If we have no possible hits
-   * fall back to secondary matching. 
-   */
-  for (a = views; a; a = a->next)
+  /* Loop over every possible desktop file that could match the window, and try
+   * to reuse an already-opened window that uses it.
+   * Desktop files are ordered by priority, so we try to use the first possible,
+   * wm_class matching applications have the priority, btw. */
+  if (possible_apps)
     {
-      view = a->data;
-
-      if (!BAMF_IS_APPLICATION (view))
-        continue;
-
-      app = BAMF_APPLICATION (view);
-      app_class = bamf_application_get_wmclass (app);
-      desktop_file = bamf_application_get_desktop_file (app);
-
-      if (possible_apps)
+      /* primary matching */
+      for (l = possible_apps; l; l = l->next)
         {
-          /* primary matching */
+          const gchar *desktop_file = l->data;
+          app = bamf_matcher_get_application_by_desktop_file (self, desktop_file);
 
-          for (l = possible_apps; l; l = l->next)
+          if (BAMF_IS_APPLICATION (app))
             {
-              if (g_strcmp0 (desktop_file, l->data) == 0)
+              const gchar *app_desktop_class;
+              app_desktop_class = bamf_application_get_wmclass (app);
+
+              if (target_class && app_desktop_class && strcasecmp (target_class, app_desktop_class) == 0)
                 {
-                  if (!best || !g_strcmp0 (win_class, app_class))
-                    best = app;
+                  best = app;
                   break;
                 }
+              else if (!best)
+                {
+                  best = app;
+                }
             }
         }
-      else if (desktop_file == NULL) 
+
+      /* If a "best" application has been found, we should check again if the
+       * desktop file that is going to be used is really the best one we have.
+       * To do this, we compare the window class name with the desktop class
+       * of both candidates to ensure that really is the best one.
+       * This is important to avoid that very-similar (which differ only by
+       * StartupWMClass) running desktop files, would be wrongly used to match
+       * an incompatible window. */
+      if (BAMF_IS_APPLICATION (best) && possible_apps)
         {
-          /* secondary matching */
-          
-          if (bamf_application_contains_similar_to_window (app, bamf_window) && (!best || !g_strcmp0 (win_class, app_class)))
-            best = app;
+          const gchar *best_app_desktop = bamf_application_get_desktop_file (best);
+          const gchar *best_desktop = possible_apps->data;
+
+          if (win_class_name && g_strcmp0 (best_app_desktop, best_desktop) != 0)
+            {
+              const gchar *best_app_class;
+              const gchar *best_desktop_class;
+
+              best_app_class = bamf_application_get_wmclass (best);
+              best_desktop_class = bamf_matcher_get_desktop_file_class (self, best_desktop);
+
+              /* We compare the two classes using their "distance" from the
+               * desidered class value */
+              if (best_app_class && best_desktop_class)
+                {
+                  int max_chars = strlen (win_class_name);
+                  int app_diff = strncasecmp (win_class_name, best_app_class, max_chars);
+                  int desktop_diff = strncasecmp (win_class_name, best_desktop_class, max_chars);
+
+                  if (abs (desktop_diff) < abs (app_diff))
+                    {
+                      best = bamf_matcher_get_application_by_desktop_file (self, best_desktop);
+                      app_desktop = best_desktop;
+                    }
+                }
+            }
+        }
+    }
+  else
+    {
+      /* secondary matching */
+      GList *a;
+      BamfView *view;
+      const gchar *app_desktop_class;
+
+      for (a = self->priv->views; a; a = a->next)
+        {
+          view = a->data;
+
+          if (!BAMF_IS_APPLICATION (view))
+            continue;
+
+          app = BAMF_APPLICATION (view);
+          app_desktop_class = bamf_application_get_wmclass (app);
+
+          if (bamf_application_contains_similar_to_window (app, bamf_window))
+            {
+              if (target_class && g_strcmp0 (target_class, app_desktop_class) == 0)
+                {
+                  best = app;
+                  break;
+                }
+              else if (!best)
+                {
+                  best = app;
+                }
+            }
         }
     }
 
   if (!best)
     {
-      if (possible_apps)
-        best = bamf_application_new_from_desktop_files (possible_apps);
+      if (app_desktop)
+        {
+          best = bamf_application_new_from_desktop_file (app_desktop);
+        }
+      else if (possible_apps)
+        {
+          best = bamf_application_new_from_desktop_files (possible_apps);
+        }
       else
-        best = bamf_application_new ();
+        {
+          best = bamf_application_new ();
+        }
 
-      bamf_application_set_wmclass (best, win_class);
+      bamf_application_set_wmclass (best, app_class);
 
-      bamf_matcher_register_view_stealing_ref (self, BAMF_VIEW (best));
+      if (new_application)
+        *new_application = TRUE;
+    }
+  else
+    {
+      if (new_application)
+        *new_application = FALSE;
     }
 
   g_list_free_full (possible_apps, g_free);
 
-  bamf_view_add_child (BAMF_VIEW (best), BAMF_VIEW (bamf_window));
+  return best;
 }
 
 /* Ensures that the window hint is set if a registered pid matches, and that set window hints
@@ -1760,7 +1947,7 @@ ensure_window_hint_set (BamfMatcher *self,
 {
   GArray *pids;
   GHashTable *registered_pids;
-  char *window_hint = NULL;
+  char *desktop_file_hint = NULL;
   gint i, pid;
   gpointer key;
 
@@ -1769,51 +1956,35 @@ ensure_window_hint_set (BamfMatcher *self,
 
   registered_pids = self->priv->registered_pids;
 
-  // web apps can end up forcing their parent browsers to match on them, so remove from pid table if this is the case
-  if (is_web_app_window (self, window))
+  /* Some windows such as web applications shares the pid with their parent
+   * browser so, we have to ignore them */
+  if (bamf_matcher_window_skips_hint_set (self, window))
     {
+      return;
+    }
+
+  desktop_file_hint = get_window_hint (window, _NET_WM_DESKTOP_FILE);
+
+  if (desktop_file_hint)
+    {
+      /* already set, make sure we know about this
+       * fact for future windows of this applications */
       pid = bamf_legacy_window_get_pid (window);
 
       if (pid > 0)
         {
-          gpointer key = GINT_TO_POINTER (pid);
-          const char* result = g_hash_table_lookup (registered_pids, key);
-          if (result && g_str_has_prefix (result, "/home/"))
+          key = GINT_TO_POINTER (pid);
+
+          if (!g_hash_table_lookup (registered_pids, key))
             {
-              g_hash_table_remove (registered_pids, key);
+              g_hash_table_insert (registered_pids, key, g_strdup (desktop_file_hint));
             }
         }
 
+      g_free (desktop_file_hint);
       return;
     }
 
-  window_hint = get_window_hint (window, _NET_WM_DESKTOP_FILE);
-  if (window_hint)
-    {
-      if (window_hint[0] != '\0')
-        {
-          /* already set, make sure we know about this
-           * fact for future windows of this applications */
-          pid = bamf_legacy_window_get_pid (window);
-
-          if (pid > 0)
-            {
-              key = GINT_TO_POINTER (pid);
-
-              if (!g_hash_table_lookup (registered_pids, key))
-                {
-                  g_hash_table_insert (registered_pids, key, g_strdup (window_hint));
-                }
-            }
-
-          g_free (window_hint);
-          return;
-        }
-
-      g_free (window_hint);
-      window_hint = NULL;
-    }
-
   pids = pid_parent_tree (self, bamf_legacy_window_get_pid (window));
 
   for (i = 0; i < pids->len; i++)
@@ -1821,21 +1992,22 @@ ensure_window_hint_set (BamfMatcher *self,
       pid = g_array_index (pids, gint, i);
       key = GINT_TO_POINTER (pid);
 
-      window_hint = g_hash_table_lookup (registered_pids, key);
-      if (window_hint != NULL && window_hint[0] != '\0')
+      desktop_file_hint = g_hash_table_lookup (registered_pids, key);
+      if (desktop_file_hint != NULL && desktop_file_hint[0] != '\0')
         break;
     }
 
   g_array_free (pids, TRUE);
 
-  if (window_hint)
-    set_window_hint (window, _NET_WM_DESKTOP_FILE, window_hint);
+  if (desktop_file_hint)
+    set_window_hint (window, _NET_WM_DESKTOP_FILE, desktop_file_hint);
 }
 
 static void
 handle_raw_window (BamfMatcher *self, BamfLegacyWindow *window)
 {
   BamfWindow *bamfwindow;
+  BamfApplication *bamfapplication;
 
   g_return_if_fail (BAMF_IS_MATCHER (self));
   g_return_if_fail (BAMF_IS_LEGACY_WINDOW (window));
@@ -1853,7 +2025,15 @@ handle_raw_window (BamfMatcher *self, BamfLegacyWindow *window)
   bamfwindow = bamf_window_new (window);
   bamf_matcher_register_view_stealing_ref (self, BAMF_VIEW (bamfwindow));
 
-  bamf_matcher_setup_window_state (self, bamfwindow);
+  gboolean new_app = FALSE;
+  bamfapplication = bamf_matcher_get_application_for_window (self, bamfwindow, &new_app);
+
+  if (new_app)
+    {
+      bamf_matcher_register_view_stealing_ref (self, BAMF_VIEW (bamfapplication));
+    }
+
+  bamf_view_add_child (BAMF_VIEW (bamfapplication), BAMF_VIEW (bamfwindow));
 }
 
 static void
@@ -1895,13 +2075,6 @@ get_gnome_control_center_window_hint (BamfMatcher * self, BamfLegacyWindow * win
   g_return_val_if_fail (BAMF_IS_LEGACY_WINDOW (window), NULL);
 
   role = get_window_hint (window, WM_WINDOW_ROLE);
-
-  if (role && role[0] == '\0')
-    {
-      g_free (role);
-      role = NULL;
-    }
-
   exec = g_strconcat ("gnome-control-center", (role ? " " : NULL), role, NULL);
 
   desktopFileTable = self->priv->desktop_file_table;
@@ -2013,15 +2186,11 @@ static void
 bamf_matcher_setup_indicator_state (BamfMatcher *self, BamfIndicator *indicator)
 {
   GList *possible_apps, *l;
-  GList *views, *a;
   const char *desktop_file;
   BamfApplication *app = NULL, *best = NULL;
-  BamfView *view;
 
   g_return_if_fail (BAMF_IS_MATCHER (self));
   g_return_if_fail (BAMF_IS_INDICATOR (indicator));
-  
-  views = self->priv->views;
 
   possible_apps = bamf_matcher_possible_applications_for_pid (self, bamf_indicator_get_pid (indicator));
 
@@ -2029,27 +2198,17 @@ bamf_matcher_setup_indicator_state (BamfMatcher *self, BamfIndicator *indicator)
    * matches with any of our possible hits. If so we match it. If we have no possible hits
    * fall back to secondary matching. 
    */
-  for (a = views; a && !best; a = a->next)
+  if (possible_apps)
     {
-      view = a->data;
-
-      if (!BAMF_IS_APPLICATION (view))
-        continue;
-
-      app = BAMF_APPLICATION (view);
-      desktop_file = bamf_application_get_desktop_file (app);
-      
-      if (possible_apps)
+      for (l = possible_apps; l; l = l->next)
         {
-          /* primary matching */
+          desktop_file = l->data;
+          app = bamf_matcher_get_application_by_desktop_file (self, desktop_file);
 
-          for (l = possible_apps; l; l = l->next)
+          if (BAMF_IS_APPLICATION (app))
             {
-              if (g_strcmp0 (desktop_file, l->data) == 0)
-                {
-                  best = app;
-                  break;
-                }
+              best = app;
+              break;
             }
         }
     }
@@ -2088,6 +2247,8 @@ void
 bamf_matcher_load_desktop_file (BamfMatcher * self,
                                 const char * desktop_file)
 {
+  GList *vl, *wl;
+
   g_return_if_fail (BAMF_IS_MATCHER (self));
 
   load_desktop_file_to_table (self,
@@ -2095,6 +2256,36 @@ bamf_matcher_load_desktop_file (BamfMatcher * self,
                               self->priv->desktop_file_table,
                               self->priv->desktop_id_table,
                               self->priv->desktop_class_table);
+
+  /* If an application with no .desktop file has windows that matches
+   * the new added .desktop file, then we try to re-match them. */
+  for (vl = self->priv->views; vl; vl = vl->next)
+    {
+      if (!BAMF_IS_APPLICATION (vl->data))
+        continue;
+
+      BamfApplication *app = BAMF_APPLICATION (vl->data);
+
+      if (!bamf_application_get_desktop_file (app))
+        {
+          GList *children = bamf_view_get_children (BAMF_VIEW (app));
+
+          for (wl = children; wl; wl = wl->next)
+          {
+            if (!BAMF_IS_WINDOW (wl->data))
+              continue;
+
+            BamfWindow *win = BAMF_WINDOW (wl->data);
+            GList *desktops = bamf_matcher_possible_applications_for_window (self, win, NULL);
+
+            if (g_list_find_custom (desktops, desktop_file, (GCompareFunc) g_strcmp0))
+              {
+                BamfLegacyWindow *legacy_window = bamf_window_get_window (win);
+                bamf_legacy_window_reopen (legacy_window);
+              }
+          }
+        }
+    }
 }
 
 void
@@ -2128,6 +2319,15 @@ bamf_matcher_register_desktop_file_for_pid (BamfMatcher * self,
     }
 }
 
+const char *
+bamf_matcher_get_desktop_file_class (BamfMatcher * self, const char * desktop_file)
+{
+  g_return_val_if_fail (BAMF_IS_MATCHER (self), NULL);
+  g_return_val_if_fail (desktop_file, NULL);
+
+  return g_hash_table_lookup (self->priv->desktop_class_table, desktop_file);
+}
+
 static int
 x_error_handler (Display *display, XErrorEvent *event)
 {
@@ -2232,7 +2432,6 @@ compare_windows_by_stack_order (gconstpointer a, gconstpointer b)
   return (idx_a < idx_b) ? -1 : 1;
 }
 
-
 GVariant *
 bamf_matcher_get_window_stack_for_monitor (BamfMatcher *matcher, gint monitor)
 {
@@ -2270,6 +2469,7 @@ bamf_matcher_get_window_stack_for_monitor (BamfMatcher *matcher, gint monitor)
       }
     }
 
+  g_list_free (windows);
   g_variant_builder_close (&b);
 
   return g_variant_builder_end (&b);
@@ -2279,27 +2479,15 @@ gboolean
 bamf_matcher_application_is_running (BamfMatcher *matcher,
                                      const char *application)
 {
-  const char * desktop_file;
-  GList *l;
-  BamfView *view;
-  BamfMatcherPrivate *priv;
+  BamfApplication *app;
 
   g_return_val_if_fail (BAMF_IS_MATCHER (matcher), FALSE);
 
-  priv = matcher->priv;
+  app = bamf_matcher_get_application_by_desktop_file (matcher, application);
 
-  for (l = priv->views; l; l = l->next)
+  if (BAMF_IS_APPLICATION (app))
     {
-      view = l->data;
-
-      if (!BAMF_IS_APPLICATION (view))
-        continue;
-
-      desktop_file = bamf_application_get_desktop_file (BAMF_APPLICATION (view));
-      if (g_strcmp0 (desktop_file, application) == 0)
-        {
-          return bamf_view_is_running (view);
-        }
+      return bamf_view_is_running (BAMF_VIEW (app));
     }
 
   return FALSE;
@@ -2370,27 +2558,15 @@ bamf_matcher_dbus_path_for_application (BamfMatcher *matcher,
                                         const char *application)
 {
   const char * path = "";
-  const char * desktop_file;
-  GList *l;
-  BamfView *view;
-  BamfMatcherPrivate *priv;
+  BamfApplication *app;
 
   g_return_val_if_fail (BAMF_IS_MATCHER (matcher), NULL);
 
-  priv = matcher->priv;
+  app = bamf_matcher_get_application_by_desktop_file (matcher, application);
 
-  for (l = priv->views; l; l = l->next)
+  if (BAMF_IS_APPLICATION (app))
     {
-      view = l->data;
-
-      if (!BAMF_IS_APPLICATION (view))
-        continue;
-
-      desktop_file = bamf_application_get_desktop_file (BAMF_APPLICATION (view));
-      if (g_strcmp0 (desktop_file, application) == 0)
-        {
-          path = bamf_view_get_path (view);
-        }
+      return bamf_view_get_path (BAMF_VIEW (app));
     }
 
   return path;
@@ -2423,8 +2599,8 @@ bamf_matcher_register_favorites (BamfMatcher *matcher,
       if (g_list_find_custom (priv->favorites, fav, (GCompareFunc) g_strcmp0))
         continue;
 
-      priv->favorites = g_list_prepend (priv->favorites, g_strdup (fav));
       bamf_matcher_load_desktop_file (matcher, fav);
+      priv->favorites = g_list_prepend (priv->favorites, g_strdup (fav));
     }
 
   g_signal_emit (matcher, matcher_signals[FAVORITES_CHANGED], 0);
@@ -2530,29 +2706,16 @@ bamf_matcher_xids_for_application (BamfMatcher *matcher,
 {
   GVariantBuilder b;
   GVariant *xids;
-  const char * desktop_file;
-  GList *l;
-  BamfView *view;
-  BamfMatcherPrivate *priv;
+  BamfApplication *app;
 
   g_return_val_if_fail (BAMF_IS_MATCHER (matcher), NULL);
 
-  priv = matcher->priv;
   xids = NULL;
+  app = bamf_matcher_get_application_by_desktop_file (matcher, application);
 
-  for (l = priv->views; l; l = l->next)
+  if (BAMF_IS_APPLICATION (app))
     {
-      view = l->data;
-
-      if (!BAMF_IS_APPLICATION (view))
-        continue;
-
-      desktop_file = bamf_application_get_desktop_file (BAMF_APPLICATION (view));
-      if (g_strcmp0 (desktop_file, application) == 0)
-        {
-          xids = bamf_application_get_xids (BAMF_APPLICATION (view));
-          break;
-        }
+      xids = bamf_application_get_xids (app);
     }
 
   if (!xids)
